"""Common Likelihoods.

All the heavy lifting is done by ``statsmodels`` + ``scipy.optimize``.
"""
import numpy as np
from statsmodels.base.model import GenericLikelihoodModel

# n_sig: param to be fitted
# variables which span llh space: alpha, energy
# additional params: n_bkg, signal_pdf, background_pdf


class BaseLLHNoExog(GenericLikelihoodModel):
    """Likelihood base class.

    The signal and background distributions need to be specified yet.

    Parameters:
        endog: list[arrays]
            collection of your data vectors on which the
            likelihood is evaluated (i.e. energy, distance to source),
        signal: callable
            Signal distribution
        background: callable
            Background distribution
    """
    def __init__(self, endog, exog=None, signal=None,
                 background=None, **kwds):
        """Don't care about exogenous variables, don't have them.

        ``endog`` is a collection of your data vectors on which the
        likelihood is evaluated (i.e. energy, distance to source),
        but not the parameter we would like to fit! (e.g. number of
        signal events).
        """
        if signal is not None:
            self.signal = signal
        if background is not None:
            self.background = background
        exog = np.zeros_like(endog)
        super(BaseLLHNoExog, self).__init__(endog=endog, exog=exog, **kwds)

    @classmethod
    def signal(cls, *args):
        """Signal distribution"""
        raise NotImplementedError

    @classmethod
    def background(cls, *args):
        """Background distribution"""
        raise NotImplementedError

    def nloglikeobs(self, params):
        endog = self.endog
        ll = self.nlnlike(params, endog)
        return ll

    def nlnlike(self, params, endog):
        """The actual likelihood, with parameter + data vectors.

        Implement this in your subclass.
        """
        raise NotImplementedError


class PointSourceStandardLLH(BaseLLHNoExog):
    """Point Source Standard Likelihood.

    The signal and background distributions need to be specified yet.

    Parameters:
        endog: list[arrays]
            collection of your data vectors on which the
            likelihood is evaluated (i.e. energy, distance to source),
        signal: callable
            Signal distribution
        background: callable
            Background distribution
    """
    def nlnlike(self, params, endog):
        """The standard Negative Log Likelihood."""
        n_sig = params[0]
        alpha, energy = endog

        n_tot = alpha.shape[0]

        sig = self.signal(alpha, energy)
        bkg = self.background(alpha, energy)
        nsumlogl = -np.sum(
            np.log(
                (n_sig / n_tot) * sig + ((n_tot - n_sig) / n_tot) * bkg
            )
        )
        return nsumlogl


class PointSourceExtendedLLH(BaseLLHNoExog):
    """Point Source Extended Likelihood.

    The signal and background distributions need to be specified yet.

    Parameters:
        endog: list[arrays]
            collection of your data vectors on which the
            likelihood is evaluated (i.e. energy, distance to source),
        n_bkg: integer, optional [default: 10000]
            Number of expected background events in sample.
        signal: callable
            Signal distribution
        background: callable
            Background distribution
    """
    def __init__(self, endog, n_bkg=10000, **kwds):
        """We need 1 additional parameter, the number of background events.
        """
        self.n_bkg = n_bkg
        super(PointSourceExtendedLLH, self).__init__(endog=endog, **kwds)

    def nlnlike(self, params, endog):
        """The extended Negative Log Likelihood."""
        n_sig = params[0]
        alpha, energy = endog
        n_tot = n_sig + self.n_bkg
        sig = self.signal(alpha, energy)
        bkg = self.background(alpha, energy)
        nsumlogl = -np.sum(
            np.log(
                n_sig * sig + self.n_bkg * bkg
            ) - n_tot - n_sig
        )
        return nsumlogl



#!/usr/bin/env python
# vim:set ts=4 sts=4 sw=4 et:

from setuptools import setup
from km3like import __version__


setup(
    name='km3like',
    version=__version__,
    description='Likelihood Utils',
    url='http://git.km3net.de/moritz/km3like',
    author='Moritz Lotze',
    author_email='mlotze@km3net.de',
    license='BSD-3',
    packages=['km3like', ],
    install_requires=[
        'numpy>=1.12',
        'pandas',
        'seaborn',
        'statsmodels',
        'km3pipe>=7.0',
        'km3astro',
    ]
)



"""Sampling routines, for pseudoexperiment generation.

Assumes pandas/numpy.
"""
import numpy as np

from km3astro.random import random_date, random_azimuth


class ScrambledBootstrap(object):
    """Resample a dataframe's events, and add random azimuth/time.

    ``(energy, zenith) -> (energy, zenith, azimuth, time)``

    Methods
    -------
    sample(n_evts=1000, weights=None)
        Generate random samples

    Examples
    --------
    >>> n_rows = 5
    >>> ingred = pd.DataFrame({
        'zenith': np.random.uniform(high=np.pi, size=n_rows),
        'energy': np.randomuniform(low=1, high=100, size=n_rows),
    })
    >>> scramboot = ScrambledBootstrap(ingred)
    >>> sampled_events = scramboot.sample(100)
    """
    def __init__(self, ingred_df, weight_col=None):
        """
        Parameters
        ----------
        ingred_df : pd.DataFrame-like
            the ingredient dataframe. required to have columns 'zenith'
            and 'energy'
        weight_col: string [default: None]
            Column holding the event weights (if existing).
            Must be specified in order to use weighting in the sampling later!
        """
        self._check_required_columns(ingred_df)
        self.ingred_df = ingred_df
        if weight_col is not None:
            if weight_col not in ingred_df.columns:
                raise KeyError(
                    "Column '{}' not in Dataframe!".format(weight_col))
        self.weight_col = weight_col

    def sample(self, n_evts=1000, weighted=False):
        """
        Parameters
        ----------
        n_evts : integer [default: 1000]
            Number of events to generate
        weighted: bool [default: False]
            Select input events according to their weights?
            Must have specified the weights column to init before!
        """
        wgt = None
        if weighted:
            if self.weight_col is None:
                raise ValueError(
                    'Weight column not specified! Cannot use weights!')
            wgt = self.ingred_df[self.weight_col]

        samp = self.ingred_df.sample(n_evts, weights=wgt, replace=True)
        samp['time'] = random_date(n_evts, astropy_time=False)
        samp['azimuth'] = random_azimuth(n_evts)
        return samp

    def _check_required_columns(self, df):
        required_columns = [
            'zenith',
            'energy',
        ]
        for col in required_columns:
            if col not in df.columns:
                raise KeyError(
                    "Required column '{}' not in dataframe!".format(col))


class Sampler(object):
    """Draw random samples from distributions.

    This expects subclasses of ``scipy.stats.rv_continuous`` (has an
    ``rvs()`` method).
    """
    def __init__(self, signal_pdf, background_pdf):
        if not hasattr(signal_pdf, 'rvs'):
            raise AttributeError("Signal pdf needs a `rvs` method!")
        if not hasattr(background_pdf, 'rvs'):
            raise AttributeError("Background pdf needs a `rvs` method!")
        self.sig_pdf = signal_pdf
        self.bkg_pdf = background_pdf

    def sample(self, n_signal, n_background=10000):
        sig = self.sig_pdf.rvs(size=n_signal)
        bkg = self.bkg_pdf.rvs(size=n_background)
        samp = np.concatenate((sig, bkg))
        return samp
